/****************************************************************************
** $Id: qt/qfontengine_win.cpp   3.3.2   edited Mar 30 12:03 $
**
** ???
**
** Copyright (C) 2003 Trolltech AS.  All rights reserved.
**
** This file is part of the kernel module of the Qt GUI Toolkit.
**
** Licensees holding valid Qt Enterprise Edition or Qt Professional Edition
** licenses for Windows may use this file in accordance with the Qt Commercial
** License Agreement provided with the Software.
**
** This file is not available for use under any other license without
** express written permission from the copyright holder.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/pricing.html or email sales@trolltech.com for
**   information about Qt Commercial License Agreements.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/

#include "qfontengine_p.h"
#include <qglobal.h>
#include "qt_windows.h"
#include "qapplication_p.h"

#include <qpaintdevice.h>
#include <qpainter.h>
#include <limits.h>
#include <math.h>

#include <private/qunicodetables_p.h>
#include <qbitmap.h>

#ifndef M_PI
#define M_PI 3.14159265358979
#endif

// defined in qtextengine_win.cpp
typedef void *SCRIPT_CACHE;
typedef HRESULT (WINAPI *fScriptFreeCache)( SCRIPT_CACHE *);
extern fScriptFreeCache ScriptFreeCache;


static unsigned char *getCMap( HDC hdc, bool & );
static Q_UINT16 getGlyphIndex( unsigned char *table, unsigned short unicode );


HDC   shared_dc	    = 0;		// common dc for all fonts
static HFONT shared_dc_font = 0;		// used by Windows 95/98
static HFONT stock_sysfont  = 0;

static inline HFONT systemFont()
{
    if ( stock_sysfont == 0 )
	stock_sysfont = (HFONT)GetStockObject(SYSTEM_FONT);
    return stock_sysfont;
}

// general font engine

QFontEngine::~QFontEngine()
{
    QT_WA( {
	if ( hdc ) {				// one DC per font (Win NT)
	    //SelectObject( hdc, systemFont() );
	    if ( !stockFont )
		DeleteObject( hfont );
	    if ( !paintDevice )
		ReleaseDC( 0, hdc );
	    hdc = 0;
	    hfont = 0;
	}
    } , {
	if ( hfont ) {				// shared DC (Windows 95/98)
	    if ( shared_dc_font == hfont ) {	// this is the current font
		Q_ASSERT( shared_dc != 0 );
		SelectObject( shared_dc, systemFont() );
		shared_dc_font = 0;
	    }
	    if ( !stockFont )
		DeleteObject( hfont );
	    hfont = 0;
	}
    } );
    delete [] cmap;

    // for Uniscribe
    if ( ScriptFreeCache )
	ScriptFreeCache( &script_cache );
}

// ##### get these from windows
int QFontEngine::lineThickness() const
{
    // ad hoc algorithm
    int score = fontDef.weight * fontDef.pixelSize;
    int lw = score / 700;

    // looks better with thicker line for small pointsizes
    if ( lw < 2 && score >= 1050 ) lw = 2;
    if ( lw == 0 ) lw = 1;

    return lw;
}

// ##### get these from windows
int QFontEngine::underlinePosition() const
{
    int pos = ( ( lineThickness() * 2 ) + 3 ) / 6;
    return pos ? pos : 1;
}


HDC QFontEngine::dc() const
{
    if ( hdc || (qt_winver & Qt::WV_NT_based) ) // either NT_based or Printer
	return hdc;
    Q_ASSERT( shared_dc != 0 && hfont != 0 );
    if ( shared_dc_font != hfont ) {
	SelectObject( shared_dc, hfont );
	shared_dc_font = hfont;
    }
    return shared_dc;
}

void QFontEngine::getCMap()
{
    QT_WA( {
	ttf = (bool)(tm.w.tmPitchAndFamily & TMPF_TRUETYPE);
    } , {
	ttf = (bool)(tm.a.tmPitchAndFamily & TMPF_TRUETYPE);
    } );
    HDC hdc = dc();
    SelectObject( hdc, hfont );
    bool symb = false;
    cmap = ttf ? ::getCMap( hdc, symb ) : 0;
    if ( !cmap ) {
	ttf = false;
	symb = false;
    }
    symbol = symb;
    script_cache = 0;
}

void QFontEngine::getGlyphIndexes( const QChar *ch, int numChars, glyph_t *glyphs, bool mirrored ) const
{
    if ( mirrored ) {
	if ( symbol ) {
	    while( numChars-- ) {
		*glyphs = getGlyphIndex(cmap, ch->unicode() );
		if(!*glyphs && ch->unicode() < 0x100)
		    *glyphs = getGlyphIndex(cmap, ch->unicode()+0xf000 );
		glyphs++;
		ch++;
	    }
	} else if ( ttf ) {
	    while( numChars-- ) {
		*glyphs = getGlyphIndex(cmap, ::mirroredChar(*ch).unicode() );
		glyphs++;
		ch++;
	    }
	} else {
	    while( numChars-- ) {
		*glyphs = ::mirroredChar(*ch).unicode();
		glyphs++;
		ch++;
	    }
	}
    } else {
	if ( symbol ) {
	    while( numChars-- ) {
		*glyphs = getGlyphIndex(cmap, ch->unicode() );
		if(!*glyphs && ch->unicode() < 0x100)
		    *glyphs = getGlyphIndex(cmap, ch->unicode()+0xf000 );
		glyphs++;
		ch++;
	    }
	} else if ( ttf ) {
	    while( numChars-- ) {
		*glyphs = getGlyphIndex(cmap, ch->unicode() );
		glyphs++;
		ch++;
	    }
	} else {
	    while( numChars-- ) {
		*glyphs = ch->unicode();
		glyphs++;
		ch++;
	    }
	}
    }
}


// non Uniscribe engine

QFontEngineWin::QFontEngineWin( const char * name, HDC _hdc, HFONT _hfont, bool stockFont, LOGFONT lf )
{
    paintDevice = FALSE;
    //qDebug("regular windows font engine created: font='%s', size=%d", name, lf.lfHeight);

    _name = name;

    hdc = _hdc;
    hfont = _hfont;
    logfont = lf;
    SelectObject( dc(), hfont );
    this->stockFont = stockFont;

    lbearing = SHRT_MIN;
    rbearing = SHRT_MIN;

    BOOL res;
    QT_WA( {
	res = GetTextMetricsW( dc(), &tm.w );
    } , {
	res = GetTextMetricsA( dc(), &tm.a );
    } );
#ifndef QT_NO_DEBUG
    if ( !res )
	qSystemWarning( "QFontPrivate: GetTextMetrics failed" );
#endif
    cache_cost = tm.w.tmHeight * tm.w.tmAveCharWidth * 2000;
    getCMap();

    useTextOutA = FALSE;
#ifndef Q_OS_TEMP
    // TextOutW doesn't work for symbol fonts on Windows 95!
    // since we're using glyph indices we don't care for ttfs about this!
    if ( qt_winver == Qt::WV_95 && !ttf &&
	 ( _name == "Marlett" || _name == "Symbol" || _name == "Webdings" || _name == "Wingdings" ) )
	    useTextOutA = TRUE;
#endif
    memset( widthCache, 0, sizeof(widthCache) );
}


QFontEngine::Error QFontEngineWin::stringToCMap( const QChar *str, int len, glyph_t *glyphs, advance_t *advances, int *nglyphs, bool mirrored ) const
{
    if ( *nglyphs < len ) {
	*nglyphs = len;
	return OutOfMemory;
    }

    getGlyphIndexes( str, len, glyphs, mirrored );

    if ( advances ) {
	HDC hdc = dc();
	unsigned int glyph;
	int overhang = (qt_winver & Qt::WV_DOS_based) ? tm.a.tmOverhang : 0;
	for( register int i = 0; i < len; i++ ) {
	    glyph = *(glyphs + i);
	    advances[i] = (glyph < widthCacheSize) ? widthCache[glyph] : 0;
	    // font-width cache failed
	    if ( !advances[i] ) {
		SIZE size;
		GetTextExtentPoint32W( hdc, (wchar_t *)str, 1, &size );
		advances[i] = size.cx - overhang;
		// if glyph's within cache range, store it for later
		if ( glyph < widthCacheSize && advances[i] > 0 && advances[i] < 0x100 )
		    ((QFontEngineWin *)this)->widthCache[glyph] = advances[i];
	    }
 	    str++;
	}
    }

    *nglyphs = len;
    return NoError;
}

#define COLOR_VALUE(c) ((p->flags & QPainter::RGBColor) ? RGB(c.red(),c.green(),c.blue()) : c.pixel())


void QFontEngineWin::draw( QPainter *p, int x, int y, const QTextEngine *engine, const QScriptItem *si, int textFlags )
{
    bool nat_xf = (qt_winver & Qt::WV_NT_based) && p->txop >= QPainter::TxScale;

    bool force_bitmap = p->rop != QPainter::CopyROP;
    force_bitmap |= p->txop >= QPainter::TxScale
		    && !(QT_WA_INLINE( tm.w.tmPitchAndFamily, tm.a.tmPitchAndFamily ) & (TMPF_VECTOR|TMPF_TRUETYPE));

    double scale = 1.;
    int angle = 0;
    bool transform = FALSE;

    if ( force_bitmap || (p->txop >= QPainter::TxScale && !nat_xf) ) {
	// Draw rotated and sheared text on Windows 95, 98

	// All versions can draw rotated text natively. Scaling can be done with window/viewport transformations
	// the hard part is only shearing

	if ( force_bitmap || p->m11() != p->m22() || p->m12() != -p->m21() ) {
	    // shearing transformation, have to do the work by hand
            QRect bbox( 0, 0, si->width, si->ascent + si->descent + 1 );
            int w=bbox.width(), h=bbox.height();
            int aw = w, ah = h;
            int tx=-bbox.x(),  ty=-bbox.y();    // text position
            QWMatrix mat1 = p->xmat;
	    if ( aw == 0 || ah == 0 )
		return;
	    double rx = (double)w / (double)aw;
	    double ry = (double)h / (double)ah;
            QWMatrix mat2 = QPixmap::trueMatrix( QWMatrix( rx, 0, 0, ry, 0, 0 )*mat1, aw, ah );

	    QBitmap bm( aw, ah, TRUE );
	    QPainter paint;
	    paint.begin( &bm );             // draw text in bitmap
	    HDC oldDC = hdc;
	    hdc = paint.handle();
	    SelectObject( hdc, hfont );
	    draw( &paint, 0, si->ascent, engine, si, textFlags );
	    hdc = oldDC;
	    paint.end();
	    QBitmap wx_bm = bm.xForm(mat2); // transform bitmap
	    if (wx_bm.isNull())
		return;

            double fx=x, fy = y - si->ascent, nfx, nfy;
            mat1.map( fx,fy, &nfx,&nfy );
            double tfx=tx, tfy=ty, dx, dy;
            mat2.map( tfx, tfy, &dx, &dy );     // compute position of bitmap
            x = qRound(nfx-dx);
            y = qRound(nfy-dy);
#if 0
	    if ( p->testf(QPainter::ExtDev) ) {		// to printer
		QRegion reg( *wx_bm );
		reg.translate( x, y );
		HBRUSH brush = CreateSolidBrush( COLOR_VALUE(p->cpen.data->color) );
		FillRgn( hdc, reg.handle(), brush );
		DeleteObject( brush );
	    } else
#endif
	    {				// to screen/pixmap
		// this code is also used in bitBlt() in qpaintdevice_win.cpp
		// (for the case that you have a selfmask)
		const DWORD ropCodes[] = {
		    0x00b8074a, // PSDPxax,  CopyROP,
		    0x00ba0b09, // DPSnao,   OrROP,
		    0x009a0709, // DPSnax,   XorROP,
		    0x008a0e06, // DSPnoa,   EraseROP=NotAndROP,
		    0x008b0666, // DSPDxoxn, NotCopyROP,
		    0x00ab0889, // DPSono,   NotOrROP,
		    0x00a90189, // DPSoxn,   NotXorROP,
		    0x00a803a9, // DPSoa,    NotEraseROP=AndROP,
		    0x00990066, // DSxn,     NotROP,
		    0x008800c6, // DSa,      ClearROP,
		    0x00bb0226, // DSno,     SetROP,
		    0x00aa0029, // D,        NopROP,
		    0x00981888, // SDPSonoxn,AndNotROP,
		    0x00b906e6, // DSPDaoxn, OrNotROP,
		    0x009b07a8, // SDPSoaxn, NandROP,
		    0x00891b08  // SDPSnaoxn,NorROP,
		};
		HBRUSH b = CreateSolidBrush( COLOR_VALUE(p->cpen.data->color) );
		COLORREF tc, bc;
		b = (HBRUSH)SelectObject( hdc, b );
		tc = SetTextColor( hdc, COLOR_VALUE(Qt::black) );
		bc = SetBkColor( hdc, COLOR_VALUE(Qt::white) );
		HDC wx_dc;
		int wx_sy;
		if ( wx_bm.isMultiCellPixmap() ) {
		    wx_dc = wx_bm.multiCellHandle();
		    wx_sy = wx_bm.multiCellOffset();
		} else {
		    wx_dc = wx_bm.handle();
		    wx_sy = 0;
		}
		BitBlt( hdc, x, y, wx_bm.width(), wx_bm.height(),
			wx_dc, 0, wx_sy, ropCodes[p->rop] );
		SetBkColor( hdc, bc );
		SetTextColor( hdc, tc );
		DeleteObject( SelectObject(hdc, b) );
	    }
            return;
	}

	// rotation + scale + translation
	scale = sqrt( p->m11()*p->m22() - p->m12()*p->m21() );
	angle = (int)(1800*acos( p->m11()/scale )/M_PI);
	if ( p->m12() < 0 )
	    angle = 3600 - angle;

	transform = TRUE;
    } else if ( nat_xf ) {
	if( !p->nativeXForm( TRUE ) ) {
 	    p->nativeXForm( FALSE );
	    return;
	}
    } else if ( p->txop == QPainter::TxTranslate ) {
	p->map( x, y, &x, &y );
    }

    if ( textFlags & Qt::Underline || textFlags & Qt::StrikeOut || scale != 1. || angle ) {
	LOGFONT lf = logfont;
	lf.lfUnderline = (bool)(textFlags & Qt::Underline);
	lf.lfStrikeOut = (bool)(textFlags & Qt::StrikeOut);
	if ( angle ) {
	    lf.lfOrientation = -angle;
	    lf.lfEscapement = -angle;
	}
	if ( scale != 1. ) {
	    lf.lfHeight = qRound(lf.lfHeight*scale);
	    lf.lfWidth = qRound(lf.lfWidth*scale);
	}
	HFONT hf = QT_WA_INLINE( CreateFontIndirectW( &lf ), CreateFontIndirectA( (LOGFONTA*)&lf ) );
	SelectObject( hdc, hf );
    }

    unsigned int options =  ttf ? ETO_GLYPH_INDEX : 0;
    if (p->backgroundMode() == Qt::OpaqueMode)
	options |= ETO_OPAQUE;

    glyph_t *glyphs = engine->glyphs( si );
    advance_t *advances = engine->advances( si );
    qoffset_t *offsets = engine->offsets( si );

    if(p->pdev->devType() == QInternal::Printer) {
	// some buggy printer drivers can't handle glyph indices correctly for latin1
	// If the string is pure latin1, we output the string directly, not the glyph indices.
	// There must be a better way to get this working, but currently I can't think of one.
	const QChar *uc = engine->string.unicode() + si->position;
	int l = engine->length(si - &engine->items[0]);
	int i = 0;
	bool latin = (l == si->num_glyphs);
	while (latin && i < l) {
	    if(uc[i].unicode() >= 0x100)
		latin = FALSE;
	    ++i;
	}
	if(latin) {
	    glyphs = (glyph_t *)uc;
	    options = 0;
	}
    }

    int xo = x;

    if ( !(si->analysis.bidiLevel % 2) ) {
	// hack to get symbol fonts working on Win95. See also QFontEngine constructor
#ifndef Q_OS_TEMP
	if ( useTextOutA ) {
	    // can only happen if !ttf
	    for( int i = 0; i < si->num_glyphs; i++ ) {
    		QChar chr = *glyphs;
		QConstString str( &chr, 1 );
		QCString cstr = str.string().local8Bit();
		TextOutA( hdc, x + offsets->x, y + offsets->y, cstr.data(), cstr.length() );
		x += *advances;
		glyphs++;
		offsets++;
		advances++;
	    }
	} else
#endif
	{
	    bool haveOffsets = si->complex | si->hasPositioning;
	    int w = 0;
            if(!haveOffsets) {
	        for( int i = 0; i < si->num_glyphs; i++ ) {
		    if ( offsets[i].x || offsets[i].y ) {
		        haveOffsets = TRUE;
		        break;
		    }
		    w += advances[i];
	        }
            }

	    if ( haveOffsets || transform ) {
		for( int i = 0; i < si->num_glyphs; i++ ) {
    		    wchar_t chr = *glyphs;
		    int xp = x + offsets->x;
		    int yp = y + offsets->y;
		    if ( transform )
			p->map( xp, yp, &xp, &yp );
    		    ExtTextOutW( hdc, xp, yp, options, 0, &chr, 1, 0 );
		    x += *advances;
		    glyphs++;
		    offsets++;
		    advances++;
		}
	    } else {
		// fast path
		int rleft = x + offsets->x;
		int rtop = y + offsets->y;
		RECT r = { x + offsets->x, y + offsets->y, rleft + w, rtop + tm.w.tmDescent + 1 };
		// Work around some Windows bugs:
		//
		// Some windows machines seem to have problems processing the ExtTextOutW call correctly.
		// On Czech Win98 bitmap fonts have wrong spacings, so we fall back to TextOut for bitmapped fonts
		// on Dos based windows.
		// Some drivers don't paint anything on Win98 using ExtTextOut, but at least return false indicating this.
		// Also in this case fall back to TextOut.
                //
                // The advances field in ExtTextOut is 0, since we handle the cases where we have explicit
                // positioning in the code path above. Not specifying advances here gives a better output
                // when using transformations.
		bool res = false;
		if(ttf || (qt_winver & Qt::WV_NT_based))
		    res = ExtTextOutW( hdc, x + offsets->x, y + offsets->y, options, &r, (wchar_t *)glyphs, si->num_glyphs, 0);
		if(!res) {
		    const QChar *ch = engine->string.unicode() + si->position;
		    int l = engine->length(si-&engine->items[0]);
		    TextOutW(hdc, x + offsets->x, y + offsets->y, (wchar_t *)ch, l);
		}
		x += w;
	    }
	}
    } else {
	offsets += si->num_glyphs;
	advances += si->num_glyphs;
	glyphs += si->num_glyphs;
	for( int i = 0; i < si->num_glyphs; i++ ) {
	    glyphs--;
	    offsets--;
	    advances--;
    	    wchar_t chr = *glyphs;
	    int xp = x + offsets->x;
	    int yp = y + offsets->y;
	    if ( transform )
		p->map( xp, yp, &xp, &yp );
	    RECT r = { xp, yp, xp + *advances, yp + tm.w.tmDescent + 1 };
    	    ExtTextOutW( hdc, xp, yp, options, &r, &chr, 1, 0 );
	    x += *advances;
	}
    }

    if ( textFlags & Qt::Underline || textFlags & Qt::StrikeOut || scale != 1. || angle )
	DeleteObject( SelectObject( hdc, hfont ) );

    if ( textFlags & Qt::Overline ) {
	int lw = lineThickness();
	int yp = y - ascent() -1;
	Rectangle( hdc, xo, yp, x, yp + lw );

    }

    if ( nat_xf )
	p->nativeXForm( FALSE );

}

glyph_metrics_t QFontEngineWin::boundingBox( const glyph_t *glyphs,
				const advance_t *advances, const qoffset_t *offsets, int numGlyphs )
{
    Q_UNUSED( glyphs );
    Q_UNUSED( offsets );

    if ( numGlyphs == 0 )
	return glyph_metrics_t();

    int w = 0;
    const advance_t *end = advances + numGlyphs;
    while( end > advances )
	w += *(--end);

    return glyph_metrics_t(0, -tm.w.tmAscent, w, tm.w.tmHeight, w, 0 );
}

glyph_metrics_t QFontEngineWin::boundingBox( glyph_t glyph )
{
#ifndef Q_OS_TEMP
    GLYPHMETRICS gm;

    if( !ttf ) {
	SIZE s;
	WCHAR ch = glyph;
	BOOL res = GetTextExtentPoint32W( dc(), &ch, 1, &s );
	Q_UNUSED( res );
	int overhang = (qt_winver & Qt::WV_DOS_based) ? tm.a.tmOverhang : 0;
	return glyph_metrics_t( 0, -tm.a.tmAscent, s.cx, tm.a.tmHeight, s.cx-overhang, 0 );
    } else {
	DWORD res = 0;
	MAT2 mat;
	mat.eM11.value = mat.eM22.value = 1;
	mat.eM11.fract = mat.eM22.fract = 0;
	mat.eM21.value = mat.eM12.value = 0;
	mat.eM21.fract = mat.eM12.fract = 0;
	QT_WA( {
	    res = GetGlyphOutlineW( dc(), glyph, GGO_METRICS|GGO_GLYPH_INDEX, &gm, 0, 0, &mat );
	} , {
	    res = GetGlyphOutlineA( dc(), glyph, GGO_METRICS|GGO_GLYPH_INDEX, &gm, 0, 0, &mat );
	} );
	if ( res != GDI_ERROR )
	    return glyph_metrics_t( gm.gmptGlyphOrigin.x, -gm.gmptGlyphOrigin.y,
				  gm.gmBlackBoxX, gm.gmBlackBoxY, gm.gmCellIncX, gm.gmCellIncY );
    }
#endif
    return glyph_metrics_t();
}

int QFontEngineWin::ascent() const
{
    return tm.w.tmAscent;
}

int QFontEngineWin::descent() const
{
    return tm.w.tmDescent;
}

int QFontEngineWin::leading() const
{
    return tm.w.tmExternalLeading;
}

int QFontEngineWin::maxCharWidth() const
{
    return tm.w.tmMaxCharWidth;
}

enum { max_font_count = 256 };
static const ushort char_table[] = {
	40,
	67,
	70,
	75,
	86,
	88,
	89,
	91,
	102,
	114,
	124,
	127,
	205,
	645,
	884,
	922,
	1070,
	12386,
	0
};

static const int char_table_entries = sizeof(char_table)/sizeof(ushort);


int QFontEngineWin::minLeftBearing() const
{
    if ( lbearing == SHRT_MIN )
	minRightBearing(); // calculates both

    return lbearing;
}

int QFontEngineWin::minRightBearing() const
{
#ifdef Q_OS_TEMP
	return 0;
#else
    if ( rbearing == SHRT_MIN ) {
	int ml = 0;
	int mr = 0;
	if ( ttf ) {
	    HDC hdc = dc();
	    SelectObject( hdc, hfont );
	    ABC *abc = 0;
	    int n = QT_WA_INLINE( tm.w.tmLastChar - tm.w.tmFirstChar, tm.a.tmLastChar - tm.a.tmFirstChar );
	    if ( n <= max_font_count ) {
		abc = new ABC[n+1];
		QT_WA( {
		    GetCharABCWidths(hdc, tm.w.tmFirstChar, tm.w.tmLastChar, abc);
		}, {
		    GetCharABCWidthsA(hdc,tm.a.tmFirstChar,tm.a.tmLastChar,abc);
		} );
	    } else {
		abc = new ABC[char_table_entries+1];
		QT_WA( {
		    for( int i = 0; i < char_table_entries; i++ )
			GetCharABCWidths(hdc, char_table[i], char_table[i], abc+i);
		}, {
		    for( int i = 0; i < char_table_entries; i++ ) {
			QCString w = QString(QChar(char_table[i])).local8Bit();
			if ( w.length() == 1 ) {
			    uint ch8 = (uchar)w[0];
			    GetCharABCWidthsA(hdc, ch8, ch8, abc+i );
			}
		    }
		} );
		n = char_table_entries;
	    }
	    ml = abc[0].abcA;
	    mr = abc[0].abcC;
    	    for ( int i = 1; i < n; i++ ) {
		if ( abc[i].abcA + abc[i].abcB + abc[i].abcC != 0 ) {
		    ml = QMIN(ml,abc[i].abcA);
		    mr = QMIN(mr,abc[i].abcC);
		}
	    }
	    delete [] abc;
	} else {
	    QT_WA( {
		ABCFLOAT *abc = 0;
		int n = tm.w.tmLastChar - tm.w.tmFirstChar+1;
		if ( n <= max_font_count ) {
		    abc = new ABCFLOAT[n];
		    GetCharABCWidthsFloat(hdc, tm.w.tmFirstChar, tm.w.tmLastChar, abc);
		} else {
		    abc = new ABCFLOAT[char_table_entries];
		    for( int i = 0; i < char_table_entries; i++ )
			GetCharABCWidthsFloat(hdc, char_table[i], char_table[i], abc+i);
		    n = char_table_entries;
		}
		float fml = abc[0].abcfA;
		float fmr = abc[0].abcfC;
		for (int i=1; i<n; i++) {
		    if ( abc[i].abcfA + abc[i].abcfB + abc[i].abcfC != 0 ) {
			fml = QMIN(fml,abc[i].abcfA);
			fmr = QMIN(fmr,abc[i].abcfC);
		    }
		}
		ml = int(fml-0.9999);
		mr = int(fmr-0.9999);
		delete [] abc;
	    } , {
		ml = 0;
		mr = -tm.a.tmOverhang;
	    } );
	}
	((QFontEngine *)this)->lbearing = ml;
	((QFontEngine *)this)->rbearing = mr;
    }

    return rbearing;
#endif
}


const char *QFontEngineWin::name() const
{
    return 0;
}

bool QFontEngineWin::canRender( const QChar *string,  int len )
{
    if ( symbol ) {
	while( len-- ) {
	    if ( getGlyphIndex( cmap, string->unicode() ) == 0 ) {
		if( string->unicode() < 0x100 ) {
		    if(getGlyphIndex( cmap, string->unicode()+0xf000) == 0)
    			return FALSE;
		} else {
		    return FALSE;
		}
	    }
	    string++;
	}
    } else if ( ttf ) {
	while( len-- ) {
	    if ( getGlyphIndex( cmap, string->unicode() ) == 0 )
		return FALSE;
	    string++;
	}
    } else {
	QT_WA( {
	    while( len-- ) {
		if ( tm.w.tmFirstChar > string[len].unicode() || tm.w.tmLastChar < string[len].unicode() )
		    return FALSE;
	    }
	}, {
	    while( len-- ) {
		if ( tm.a.tmFirstChar > string[len].unicode() || tm.a.tmLastChar < string[len].unicode() )
		    return FALSE;
	    }
	} );
    }
    return TRUE;
}

QFontEngine::Type QFontEngineWin::type() const
{
    return QFontEngine::Win;
}


// Uniscribe engine
#if 0
typedef HRESULT (WINAPI *fScriptTextOut)( const HDC, SCRIPT_CACHE *, int, int, UINT, const RECT *, const QScriptAnalysis *,
					 const WCHAR *, int, const WORD *, int, const int *, const int *, const GOFFSET *);
extern fScriptTextOut ScriptTextOut;

void QFontEngineUniscribe::draw( QPainter *p, int x, int y, const glyph_t *glyphs,
	   const advance_t *advances, const qoffset_t *offsets, int numGlyphs, bool reverse, int textFlags )
{
    ScriptTextOut(
    HDC hdc = dc();
    ScriptTextOut( hdc, script_cache, x, y, 0, 0, analysis, 0, 0, glyphs, numGlyphs, advances, 0, offsets );
}

bool QFontEngineUniscribe::canRender( const QChar *string,  int len )
{
    // ### get default glyph for script and then check.
    return TRUE;
}

QFontEngine::Type QFontEngineUniscribe::type() const
{
    return QFontEngine::Uniscribe;
}

#endif



// box font engine



QFontEngineBox::QFontEngineBox( int size )
    : _size( size )
{
    cache_cost = 1;
    hdc = (qt_winver & Qt::WV_NT_based) ? GetDC( 0 ) : shared_dc;
#ifndef Q_OS_TEMP
    hfont = (HFONT)GetStockObject( ANSI_VAR_FONT );
#endif
    stockFont = TRUE;
    paintDevice = FALSE;
    ttf = FALSE;

    cmap = 0;
    script_cache = 0;

    //qDebug("box font engine created!");
}

QFontEngineBox::~QFontEngineBox()
{
}

QFontEngine::Error QFontEngineBox::stringToCMap( const QChar *,  int len, glyph_t *glyphs, advance_t *advances, int *nglyphs, bool ) const
{
    if ( *nglyphs < len ) {
	*nglyphs = len;
	return OutOfMemory;
    }

    for ( int i = 0; i < len; i++ )
	*(glyphs++) = 0;
    *nglyphs = len;

    if ( advances ) {
	for ( int i = 0; i < len; i++ )
	    *(advances++) = _size;
    }
    return NoError;
}

void QFontEngineBox::draw( QPainter *p, int x, int y, const QTextEngine *engine, const QScriptItem *si, int textFlags )
{
    Q_UNUSED( p );
    Q_UNUSED( x );
    Q_UNUSED( y );
    Q_UNUSED( engine );
    Q_UNUSED( si );
    Q_UNUSED( textFlags );
//     qDebug("QFontEngineXLFD::draw( %d, %d, numglyphs=%d", x, y, numGlyphs );

    // ########
}

glyph_metrics_t QFontEngineBox::boundingBox( const glyph_t *, const advance_t *, const qoffset_t *, int numGlyphs )
{
    glyph_metrics_t overall;
    overall.x = overall.y = 0;
    overall.width = _size*numGlyphs;
    overall.height = _size;
    overall.xoff = overall.width;
    overall.yoff = 0;
    return overall;
}

glyph_metrics_t QFontEngineBox::boundingBox( glyph_t )
{
    return glyph_metrics_t( 0, _size, _size, _size, _size, 0 );
}



int QFontEngineBox::ascent() const
{
    return _size;
}

int QFontEngineBox::descent() const
{
    return 0;
}

int QFontEngineBox::leading() const
{
    int l = qRound( _size * 0.15 );
    return (l > 0) ? l : 1;
}

int QFontEngineBox::maxCharWidth() const
{
    return _size;
}

const char *QFontEngineBox::name() const
{
    return "null";
}

bool QFontEngineBox::canRender( const QChar *,  int )
{
    return TRUE;
}

QFontEngine::Type QFontEngineBox::type() const
{
    return Box;
}





// ----------------------------------------------------------------------------
// True type support methods
// ----------------------------------------------------------------------------




#define MAKE_TAG(ch1, ch2, ch3, ch4) (\
    (((DWORD)(ch4)) << 24) | \
    (((DWORD)(ch3)) << 16) | \
    (((DWORD)(ch2)) << 8) | \
    ((DWORD)(ch1)) \
    )

static inline Q_UINT32 getUInt(unsigned char *p)
{
    Q_UINT32 val;
    val = *p++ << 24;
    val |= *p++ << 16;
    val |= *p++ << 8;
    val |= *p;

    return val;
}

static inline Q_UINT16 getUShort(unsigned char *p)
{
    Q_UINT16 val;
    val = *p++ << 8;
    val |= *p;

    return val;
}

static inline void tag_to_string( char *string, Q_UINT32 tag )
{
    string[0] = (tag >> 24)&0xff;
    string[1] = (tag >> 16)&0xff;
    string[2] = (tag >> 8)&0xff;
    string[3] = tag&0xff;
    string[4] = 0;
}

static Q_UINT16 getGlyphIndex( unsigned char *table, unsigned short unicode )
{
    unsigned short format = getUShort( table );
    if ( format == 0 ) {
	if ( unicode < 256 )
	    return (int) *(table+6+unicode);
    } else if ( format == 2 ) {
	qWarning("format 2 encoding table for Unicode, not implemented!");
    } else if ( format == 4 ) {
	Q_UINT16 segCountX2 = getUShort( table + 6 );
	unsigned char *ends = table + 14;
	Q_UINT16 endIndex = 0;
	int i = 0;
	for ( ; i < segCountX2/2 && (endIndex = getUShort( ends + 2*i )) < unicode; i++ );

	unsigned char *idx = ends + segCountX2 + 2 + 2*i;
	Q_UINT16 startIndex = getUShort( idx );

	if ( startIndex > unicode )
	    return 0;

	idx += segCountX2;
	Q_INT16 idDelta = (Q_INT16)getUShort( idx );
	idx += segCountX2;
	Q_UINT16 idRangeoffset_t = (Q_UINT16)getUShort( idx );

	Q_UINT16 glyphIndex;
	if ( idRangeoffset_t ) {
	    Q_UINT16 id = getUShort( idRangeoffset_t + 2*(unicode - startIndex) + idx);
	    if ( id )
		glyphIndex = ( idDelta + id ) % 0x10000;
	    else
		glyphIndex = 0;
	} else {
	    glyphIndex = (idDelta + unicode) % 0x10000;
	}
	return glyphIndex;
    }

    return 0;
}


static unsigned char *getCMap( HDC hdc, bool &symbol )
{
    const DWORD CMAP = MAKE_TAG( 'c', 'm', 'a', 'p' );

    unsigned char header[4];

    // get the CMAP header and the number of encoding tables
    DWORD bytes =
#ifndef Q_OS_TEMP
	GetFontData( hdc, CMAP, 0, &header, 4 );
#else
	0;
#endif
    if ( bytes == GDI_ERROR )
	return 0;
    unsigned short version = getUShort( header );
    if ( version != 0 )
	return 0;

    unsigned short numTables = getUShort( header+2);
    unsigned char *maps = new unsigned char[8*numTables];

    // get the encoding table and look for Unicode
#ifndef Q_OS_TEMP
    bytes = GetFontData( hdc, CMAP, 4, maps, 8*numTables );
#endif
    if ( bytes == GDI_ERROR )
	return 0;

    symbol = TRUE;
    unsigned int unicode_table = 0;
    for ( int n = 0; n < numTables; n++ ) {
	Q_UINT32 version = getUInt( maps + 8*n );
	// accept both symbol and Unicode encodings. prefer unicode.
	if ( version == 0x00030001 || version == 0x00030000 ) {
	    unicode_table = getUInt( maps + 8*n + 4 );
	    if ( version == 0x00030001 ) {
		symbol = FALSE;
		break;
	    }
	}
    }

    if ( !unicode_table ) {
	// qDebug("no unicode table found" );
	return 0;
    }

    delete [] maps;

    // get the header of the unicode table
#ifndef Q_OS_TEMP
    bytes = GetFontData( hdc, CMAP, unicode_table, &header, 4 );
#endif
    if ( bytes == GDI_ERROR )
	return 0;

    unsigned short length = getUShort( header+2 );
    unsigned char *unicode_data = new unsigned char[length];

    // get the cmap table itself
#ifndef Q_OS_TEMP
    bytes = GetFontData( hdc, CMAP, unicode_table, unicode_data, length );
#endif
    if ( bytes == GDI_ERROR ) {
	delete [] unicode_data;
	return 0;
    }
    return unicode_data;
}



